@@ -17,7 +17,7 @@
 import cliapp
 
 
-__version__ = '0.25'
+__version__ = '1.1'
 
 
 
@@ -60,6 +60,13 @@ IDPATH_SKIP = 13
 # for on-disk data structures.
 MAX_ID = 2**64 - 1
 
+
+option_group = {
+    'perf': 'Performance tweaking',
+    'devel': 'Development of Obnam itself',
+}
+
+
 from sizeparse import SizeSyntaxError, UnitNameError, ByteSizeParser
 
 from encryption import (generate_symmetric_key,
@@ -72,7 +79,7 @@ from encryption import (generate_symmetric_key,
                         decrypt_with_secret_keys,
                         SymmetricKeyCache)
 
-from hooks import Hook, FilterHook, HookManager
+from hooks import Hook, MissingFilterError, FilterHook, HookManager
 from pluginbase import ObnamPlugin
 from vfs import VirtualFileSystem, VfsFactory, VfsTests
 from vfs_local import LocalFS
@@ -83,6 +90,7 @@ from chunklist import ChunkList
 from clientlist import ClientList
 from checksumtree import ChecksumTree
 from clientmetadatatree import ClientMetadataTree
+from lockmgr import LockManager
 from repo import Repository, LockFail, BadFormat
 from forget_policy import ForgetPolicy
 from app import App
