@@ -17,6 +17,7 @@
 import hashlib
 import logging
 import os
+import random
 import struct
 import tracing
 
@@ -42,6 +43,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
     FILE_CHUNKS = 1         # subkey type for list of chunks
     FILE_METADATA = 3       # subkey type for inode fields, etc
     DIR_CONTENTS = 4        # subkey type for list of directory contents
+    FILE_DATA = 5           # subkey type for file data (instead of chunk)
     
     FILE_METADATA_ENCODED = 0 # subkey value for encoded obnamlib.Metadata().
     
@@ -70,16 +72,20 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
                  repo):
         tracing.trace('new ClientMetadataTree, client_dir=%s' % client_dir)
         self.current_time = repo.current_time
-        key_bytes = len(self.hashkey(0, self.hash_name(''), 0, 0))
+        key_bytes = len(self.hashkey(0, self.default_file_id(''), 0, 0))
         obnamlib.RepositoryTree.__init__(self, fs, client_dir, key_bytes, 
                                          node_size, upload_queue_size, 
                                          lru_size, repo)
-        self.genhash = self.hash_name('generation')
-        self.known_generations = dict()
+        self.genhash = self.default_file_id('generation')
         self.chunkids_per_key = max(1,
                                     int(node_size / 4 / struct.calcsize('Q')))
+        self.init_caches()
 
-    def hash_name(self, filename):
+    def init_caches(self):
+        self.known_generations = {}
+        self.file_ids = {}
+
+    def default_file_id(self, filename):
         '''Return hash of filename suitable for use as main key.'''
         tracing.trace(repr(filename))
         def hash(s):
@@ -88,6 +94,10 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         basename = os.path.basename(filename)
         return hash(dirname) + hash(basename)
 
+    def _bad_default_file_id(self, filename):
+        '''For use by unit tests.'''
+        return struct.pack('!Q', 0)
+
     def hashkey(self, prefix, mainhash, subtype, subkey):
         '''Compute a full key.
 
@@ -123,6 +133,11 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
     def fskey(self, mainhash, subtype, subkey):
         ''''Generate key for filesystem metadata.'''
         return self.hashkey(self.PREFIX_FS_META, mainhash, subtype, subkey)
+        
+    def fs_unkey(self, key):
+        '''Inverse of fskey.'''
+        parts = struct.unpack('!B8sB8s', key)
+        return parts[1], parts[3]
 
     def genkey(self, subkey):
         '''Generate key for generation metadata.'''
@@ -142,12 +157,51 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         parts = struct.unpack('!BQBQ', key)
         return parts[1], parts[3]
 
-    def get_file_id(self, gen, pathname):
+    def get_file_id(self, tree, pathname):
         '''Return id for file in a given generation.'''
         
-        # FIXME: This should handle hash collisions eventually.
+        if tree in self.file_ids:
+            if pathname in self.file_ids[tree]:
+                return self.file_ids[tree][pathname]
+        else:
+            self.file_ids[tree] = {}
         
-        return self.hash_name(pathname)
+        default_file_id = self.default_file_id(pathname)
+        minkey = self.fskey(default_file_id, self.FILE_NAME, 0)
+        maxkey = self.fskey(default_file_id, self.FILE_NAME, obnamlib.MAX_ID)
+        for key, value in tree.lookup_range(minkey, maxkey):
+            def_id, file_id = self.fs_unkey(key)
+            assert def_id == default_file_id, \
+                'def=%s other=%s' % (repr(def_id), repr(default_file_id))
+            self.file_ids[tree][value] = file_id
+            if value == pathname:
+                return file_id
+
+        raise KeyError('%s does not yet have a file-id' % pathname)
+
+    def set_file_id(self, pathname):
+        '''Set and return the file-id for a file in current generation.'''
+
+        default_file_id = self.default_file_id(pathname)
+        minkey = self.fskey(default_file_id, self.FILE_NAME, 0)
+        maxkey = self.fskey(default_file_id, self.FILE_NAME, obnamlib.MAX_ID)
+        file_ids = set()
+        for key, value in self.tree.lookup_range(minkey, maxkey):
+            def_id, file_id = self.fs_unkey(key)
+            assert def_id == default_file_id
+            if value == pathname:
+                return file_id
+            file_ids.add(file_id)
+        
+        while True:
+            n = random.randint(0, obnamlib.MAX_ID)
+            file_id = struct.pack('!Q', n)
+            if file_id not in file_ids:
+                break
+        
+        key = self.fskey(default_file_id, self.FILE_NAME, file_id)
+        self.tree.insert(key, pathname)
+        return file_id
 
     def _lookup_int(self, tree, key):
         return struct.unpack('!Q', tree.lookup(key))[0]
@@ -160,28 +214,49 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         if self.tree:
             now = int(self.current_time())
             self._insert_int(self.tree, self.genkey(self.GEN_ENDED), now)
-            genid = self.get_generation_id(self.tree)
-            self._insert_count(genid, self.GEN_FILE_COUNT, self.file_count)
-            self._insert_count(genid, self.GEN_TOTAL_DATA, self.total_data)
+            genid = self._get_generation_id_or_None(self.tree)
+            if genid is not None:
+                t = [(self.GEN_FILE_COUNT, 'file_count'),
+                     (self.GEN_TOTAL_DATA, 'total_data')]
+                for subkey, attr in t:
+                    if hasattr(self, attr):
+                        self._insert_count(genid, subkey, getattr(self, attr))
         obnamlib.RepositoryTree.commit(self)
 
-    def find_generation(self, genid):
-        if genid in self.known_generations:
-            return self.known_generations[genid]
+    def init_forest(self, *args, **kwargs):
+        self.init_caches()
+        return obnamlib.RepositoryTree.init_forest(self, *args, **kwargs)
 
-        if self.forest:
+    def start_changes(self, *args, **kwargs):
+        self.init_caches()
+        return obnamlib.RepositoryTree.start_changes(self, *args, **kwargs)
+
+    def find_generation(self, genid):
+    
+        def fill_cache():
             key = self.genkey(self.GEN_ID)
             for t in self.forest.trees:
                 t_genid = self._lookup_int(t, key)
-                self.known_generations[t_genid] = t
                 if t_genid == genid:
+                    self.known_generations[genid] = t
                     return t
+    
+        if self.forest:
+            if genid in self.known_generations:
+                return self.known_generations[genid]
+            t = fill_cache()
+            if t is not None:
+                return t
         raise KeyError('Unknown generation %s' % genid)
 
     def list_generations(self):
         if self.forest:
-            key = self.genkey(self.GEN_ID)
-            return [self._lookup_int(t, key) for t in self.forest.trees]
+            genids = []
+            for t in self.forest.trees:
+                genid = self._get_generation_id_or_None(t)
+                if genid is not None:
+                    genids.append(genid)
+            return genids
         else:
             return []
 
@@ -219,6 +294,12 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
     def get_generation_id(self, tree):
         return self._lookup_int(tree, self.genkey(self.GEN_ID))
 
+    def _get_generation_id_or_None(self, tree):
+        try:
+            return self.get_generation_id(tree)
+        except KeyError: # pragma: no cover
+            return None
+
     def _lookup_time(self, tree, what):
         try:
             return self._lookup_int(tree, self.genkey(what))
@@ -251,7 +332,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
 
     def create(self, filename, encoded_metadata):
         tracing.trace('filename=%s', filename)
-        file_id = self.get_file_id(self.tree, filename)
+        file_id = self.set_file_id(filename)
         gen_id = self.get_generation_id(self.tree)
         try:
             old_metadata = self.get_metadata(gen_id, filename)
@@ -276,7 +357,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         tracing.trace('parent=%s', parent)
         if parent != filename: # root dir is its own parent
             basename = os.path.basename(filename)
-            parent_id = self.get_file_id(self.tree, parent)
+            parent_id = self.set_file_id(parent)
             key = self.fskey(parent_id, self.DIR_CONTENTS, file_id)
             # We could just insert, but that would cause unnecessary
             # churn in the tree if nothing changes.
@@ -297,7 +378,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
     def set_metadata(self, filename, encoded_metadata):
         tracing.trace('filename=%s', filename)
 
-        file_id = self.get_file_id(self.tree, filename)
+        file_id = self.set_file_id(filename)
         key1 = self.fskey(file_id, self.FILE_NAME, file_id)
         self.tree.insert(key1, filename)
         
@@ -336,18 +417,26 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         minkey = self.fskey(file_id, 0, 0)
         maxkey = self.fskey(file_id, self.TYPE_MAX, self.SUBKEY_MAX)
         self.tree.remove_range(minkey, maxkey)
-
+        
+        # Remove filename.
+        default_file_id = self.default_file_id(filename)
+        key = self.fskey(default_file_id, self.FILE_NAME, file_id)
+        self.tree.remove_range(key, key)
+        
         # Also remove from parent's contents.
         parent = os.path.dirname(filename)
         if parent != filename: # root dir is its own parent
-            parent_id = self.get_file_id(self.tree, parent)
+            parent_id = self.set_file_id(parent)
             key = self.fskey(parent_id, self.DIR_CONTENTS, file_id)
             # The range removal will work even if the key does not exist.
             self.tree.remove_range(key, key)
             
     def listdir(self, genid, dirname):
         tree = self.find_generation(genid)
-        dir_id = self.get_file_id(tree, dirname)
+        try:
+            dir_id = self.get_file_id(tree, dirname)
+        except KeyError:
+            return []
         minkey = self.fskey(dir_id, self.DIR_CONTENTS, 0)
         maxkey = self.fskey(dir_id, self.DIR_CONTENTS, self.SUBKEY_MAX)
         basenames = []
@@ -357,7 +446,10 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
 
     def get_file_chunks(self, genid, filename):
         tree = self.find_generation(genid)
-        file_id = self.get_file_id(tree, filename)
+        try:
+            file_id = self.get_file_id(tree, filename)
+        except KeyError:
+            return []
         minkey = self.fskey(file_id, self.FILE_CHUNKS, 0)
         maxkey = self.fskey(file_id, self.FILE_CHUNKS, self.SUBKEY_MAX)
         pairs = tree.lookup_range(minkey, maxkey)
@@ -385,7 +477,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         tracing.trace('filename=%s', filename)
         tracing.trace('chunkids=%s', repr(chunkids))
     
-        file_id = self.get_file_id(self.tree, filename)
+        file_id = self.set_file_id(filename)
         minkey = self.fskey(file_id, self.FILE_CHUNKS, 0)
         maxkey = self.fskey(file_id, self.FILE_CHUNKS, self.SUBKEY_MAX)
 
@@ -402,7 +494,7 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         tracing.trace('filename=%s', filename)
         tracing.trace('chunkids=%s', repr(chunkids))
 
-        file_id = self.get_file_id(self.tree, filename)
+        file_id = self.set_file_id(filename)
 
         minkey = self.fskey(file_id, self.FILE_CHUNKS, 0)
         maxkey = self.fskey(file_id, self.FILE_CHUNKS, self.SUBKEY_MAX)
@@ -433,3 +525,27 @@ class ClientMetadataTree(obnamlib.RepositoryTree):
         return list(set(self.chunk_unkey(key)[0]
                         for key, value in t.lookup_range(minkey, maxkey)))
 
+    def set_file_data(self, filename, contents): # pragma: no cover
+        '''Store contents of file, if small, in B-tree instead of chunk.
+        
+        The length of the contents should be small enough to fit in a
+        B-tree leaf.
+        
+        '''
+        tracing.trace('filename=%s' % filename)
+        tracing.trace('contents=%s' % repr(contents))
+        
+        file_id = self.set_file_id(filename)
+        key = self.fskey(file_id, self.FILE_DATA, 0)
+        self.tree.insert(key, contents)
+
+    def get_file_data(self, gen_id, filename): # pragma: no cover
+        '''Return contents of file, if set, or None.'''
+        tree = self.find_generation(gen_id)
+        file_id = self.get_file_id(tree, filename)
+        key = self.fskey(file_id, self.FILE_DATA, 0)
+        try:
+            return tree.lookup(key)
+        except KeyError:
+            return None
+
