@@ -14,11 +14,13 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
+import errno
 import gc
 import logging
 import os
 import re
 import stat
+import time
 import traceback
 import tracing
 import ttystatus
@@ -26,31 +28,66 @@ import ttystatus
 import obnamlib
 
 
+class ChunkidPool(object):
+
+    '''Checksum/chunkid mappings that are pending an upload to shared trees.'''
+    
+    def __init__(self):
+        self.clear()
+        
+    def add(self, chunkid, checksum):
+        if checksum not in self._mapping:
+            self._mapping[checksum] = []
+        self._mapping[checksum].append(chunkid)
+
+    def __contains__(self, checksum):
+        return checksum in self._mapping
+
+    def get(self, checksum):
+        return self._mapping.get(checksum, [])
+        
+    def clear(self):
+        self._mapping = {}
+        
+    def __iter__(self):
+        for checksum in self._mapping.keys():
+            for chunkid in self._mapping[checksum]:
+                yield chunkid, checksum
+
+
 class BackupPlugin(obnamlib.ObnamPlugin):
 
     def enable(self):
+        backup_group = obnamlib.option_group['backup'] = 'Backing up'
+        perf_group = obnamlib.option_group['perf']
+    
         self.app.add_subcommand('backup', self.backup,
                                 arg_synopsis='[FILE]...')
         self.app.settings.string_list(['root'], 'what to backup')
         self.app.settings.string_list(['exclude'], 
                                  'regular expression for pathnames to '
                                  'exclude from backup (can be used multiple '
-                                 'times)')
+                                 'times)',
+                                 group=backup_group)
         self.app.settings.boolean(['exclude-caches'],
                                     'exclude directories (and their subdirs) '
-                                    'that contain a CACHEDIR.TAG file')
+                                    'that contain a CACHEDIR.TAG file',
+                                 group=backup_group)
         self.app.settings.boolean(['one-file-system'],
                                     'exclude directories (and their subdirs) '
-                                    'that are in a different filesystem')
+                                    'that are in a different filesystem',
+                                 group=backup_group)
         self.app.settings.bytesize(['checkpoint'],
                                       'make a checkpoint after a given SIZE '
                                       '(%default)',
                                     metavar='SIZE',
-                                    default=1024**3)
+                                    default=1024**3,
+                                 group=backup_group)
         self.app.settings.integer(['chunkids-per-group'],
                                   'encode NUM chunk ids per group (%default)',
                                   metavar='NUM',
-                                  default=obnamlib.DEFAULT_CHUNKIDS_PER_GROUP)
+                                  default=obnamlib.DEFAULT_CHUNKIDS_PER_GROUP,
+                                  group=perf_group)
         self.app.settings.choice(['deduplicate'],
                                  ['fatalist', 'never', 'verify'],
                                  'find duplicate data in backed up data '
@@ -59,33 +96,100 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                                     'verify that no hash collisions happen, '
                                     'or (the default) fatalistically accept '
                                     'the risk of collisions',
-                                 metavar='MODE')
+                                 metavar='MODE',
+                                 group=backup_group)
         self.app.settings.boolean(['leave-checkpoints'],
                                   'leave checkpoint generations at the end '
-                                    'of a successful backup run')
+                                    'of a successful backup run',
+                                 group=backup_group)
+        self.app.settings.boolean(['small-files-in-btree'],
+                                  'put contents of small files directly into '
+                                    'the per-client B-tree, instead of '
+                                    'separate chunk files; do not use this '
+                                    'as it is quite bad for performance',
+                                 group=backup_group)
 
-    def configure_ttystatus(self):
+    def configure_ttystatus_for_backup(self):
         self.app.ts['current-file'] = ''
         self.app.ts['uploaded-bytes'] = 0
-        
-        self.app.ts.add(ttystatus.ElapsedTime())
-        self.app.ts.add(ttystatus.Literal(' '))
-        self.app.ts.add(ttystatus.Counter('current-file'))
-        self.app.ts.add(ttystatus.Literal(' files; '))
-        self.app.ts.add(ttystatus.ByteSize('uploaded-bytes'))
-        self.app.ts.add(ttystatus.Literal(' up ('))
-        self.app.ts.add(ttystatus.ByteSpeed('uploaded-bytes'))
-        self.app.ts.add(ttystatus.Literal(') '))
-        self.app.ts.add(ttystatus.String('what'))
-        self.app.ts.add(ttystatus.Pathname('current-file'))
+        self.file_count = 0
+        self.uploaded_bytes = 0
+
+        self.app.ts.format('%ElapsedTime() '
+                           '%Counter(current-file) '
+                           'files; '
+                           '%ByteSize(uploaded-bytes) '
+                           '('
+                           '%ByteSpeed(uploaded-bytes,10)'
+                           ') '
+                           '%String(what)')
+
+    def configure_ttystatus_for_checkpoint_removal(self):
+        self.app.ts['what'] = 'removing checkpoints'
 
     def update_progress_with_file(self, filename, metadata):
-        self.app.ts['what'] = ''
+        self.app.ts['what'] = filename
         self.app.ts['current-file'] = filename
+        self.file_count += 1
 
     def update_progress_with_upload(self, amount):
-        self.app.ts['what'] = ''
         self.app.ts['uploaded-bytes'] += amount
+        self.uploaded_bytes += amount
+
+    def report_stats(self):
+        size_table = [
+            (1024**4, 'TiB'),
+            (1024**3, 'GiB'),
+            (1024**2, 'MiB'),
+            (1024**1, 'KiB'),
+            (0, 'B')
+        ]
+        
+        for size_base, size_unit in size_table:
+            if self.uploaded_bytes >= size_base:
+                if size_base > 0:
+                    size_amount = float(self.uploaded_bytes) / float(size_base)
+                else:
+                    size_amount = float(self.uploaded_bytes)
+                break
+
+        speed_table = [
+            (1024**3, 'GiB/s'),
+            (1024**2, 'MiB/s'),
+            (1024**1, 'KiB/s'),
+            (0, 'B/s')
+        ]
+        duration = time.time() - self.started
+        speed = float(self.uploaded_bytes) / duration
+        for speed_base, speed_unit in speed_table:
+            if speed >= speed_base:
+                if speed_base > 0:
+                    speed_amount = speed / speed_base
+                else:
+                    speed_amount = speed
+                break
+
+        duration_string = ''
+        seconds = duration
+        if seconds >= 3600:
+            duration_string += '%dh' % int(seconds/3600)
+            seconds %= 3600
+        if seconds >= 60:
+            duration_string += '%dm' % int(seconds/60)
+            seconds %= 60
+        if seconds > 0:
+            duration_string += '%ds' % round(seconds)
+
+        logging.info('Backup performance statistics:')
+        logging.info('* files found: %s' % self.file_count)
+        logging.info('* uploaded data: %s bytes (%s %s)' % 
+                        (self.uploaded_bytes, size_amount, size_unit))
+        logging.info('* duration: %s s' % duration)
+        logging.info('* average speed: %s %s' % (speed_amount, speed_unit))
+        self.app.ts.notify('Backed up %d files, uploaded %.1f %s '
+                           'in %s at %.1f %s average speed' %
+                            (self.file_count, size_amount, size_unit,
+                             duration_string, speed_amount, speed_unit))
 
     def error(self, msg, exc=None):
         self.errors = True
@@ -110,9 +214,21 @@ class BackupPlugin(obnamlib.ObnamPlugin):
 
         self.app.settings.require('repository')
         self.app.settings.require('client-name')
+        
+        if not self.app.settings['repository']:
+            raise obnamlib.Error('No --repository setting. '
+                                  'You need to specify it on the command '
+                                  'line or a configuration file.')
+        
+        # This is ugly, but avoids having to update the dependency on
+        # ttystatus yet again.
+        if not hasattr(self.app.ts, 'flush'):
+            self.app.ts.flush = lambda: None
 
-        self.configure_ttystatus()
-        self.app.ts['what'] = 'setting up;'
+        self.started = time.time()
+        self.configure_ttystatus_for_backup()
+        self.app.ts['what'] = 'setting up'
+        self.app.ts.flush()
 
         self.compile_exclusion_patterns()
         self.memory_dump_counter = 0
@@ -125,8 +241,14 @@ class BackupPlugin(obnamlib.ObnamPlugin):
             self.repo = self.app.open_repository(create=True)
             self.add_client(client_name)
             self.repo.lock_client(client_name)
+            
+            # Need to lock the shared stuff briefly, so encryption etc
+            # gets initialized.
+            self.repo.lock_shared()
+            self.repo.unlock_shared()
 
         self.errors = False
+        self.chunkid_pool = ChunkidPool()
         try:
             if not self.pretend:
                 self.repo.start_generation()
@@ -134,27 +256,57 @@ class BackupPlugin(obnamlib.ObnamPlugin):
             roots = self.app.settings['root'] + args
             if roots:
                 self.backup_roots(roots)
-            self.app.ts['what'] = 'committing changes;'
+            self.app.ts['what'] = 'committing changes'
+            self.app.ts.flush()
             if not self.pretend:
+                self.repo.lock_shared()
+                self.add_chunks_to_shared()
                 self.repo.commit_client()
+                self.repo.commit_shared()
             self.repo.fs.close()
-            self.app.ts['what'] = ''
-            self.app.ts.finish()
+            self.app.ts.clear()
+            self.report_stats()
 
             logging.info('Backup finished.')
             self.app.dump_memory_profile('at end of backup run')
-        except BaseException:
+        except BaseException, e:
+            logging.debug('Handling exception %s' % str(e))
+            logging.debug(traceback.format_exc())
+            self.unlock_when_error()
+            raise
+
+        if self.errors:
+            raise obnamlib.Error('There were errors during the backup')
+
+    def unlock_when_error(self):
+        try:
             if self.repo.got_client_lock:
                 logging.info('Unlocking client because of error')
                 self.repo.unlock_client()
-            raise
-            
+            if self.repo.got_shared_lock:
+                logging.info('Unlocking shared trees because of error')
+                self.repo.unlock_shared()
+        except BaseException, e2:
+            logging.debug('Got second exception while unlocking: %s' % 
+                            str(e2))
+            logging.debug(traceback.format_exc())
+
+    def add_chunks_to_shared(self):
+        for chunkid, checksum in self.chunkid_pool:
+            self.repo.put_chunk_in_shared_trees(chunkid, checksum)
+        self.chunkid_pool.clear()
+
     def add_client(self, client_name):
+        self.repo.lock_root()
         if client_name not in self.repo.list_clients():
             tracing.trace('adding new client %s' % client_name)
-            self.repo.lock_root()
+            tracing.trace('client list before adding: %s' % 
+                            self.repo.list_clients())
             self.repo.add_client(client_name)
-            self.repo.commit_root()
+            tracing.trace('client list after adding: %s' % 
+                            self.repo.list_clients())
+        self.repo.commit_root()
+        self.repo = self.app.open_repository(repofs=self.repo.fs.fs)
 
     def compile_exclusion_patterns(self):
         log = self.app.settings['log']
@@ -204,12 +356,11 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                         metadata.md5 = self.backup_file_contents(pathname,
                                                                  metadata)
                     self.backup_metadata(pathname, metadata)
-                except OSError, e:
-                    msg = 'Can\'t back up %s: %s' % (pathname, e.strerror)
-                    self.error(msg, e)
-                except IOError, e:
+                except (IOError, OSError), e:
                     msg = 'Can\'t back up %s: %s' % (pathname, e.strerror)
                     self.error(msg, e)
+                    if e.errno == errno.ENOSPC:
+                        raise
                 if self.time_for_checkpoint():
                     self.make_checkpoint()
 
@@ -219,9 +370,9 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                               not self.app.settings['leave-checkpoints']
                               and not self.pretend)
         if remove_checkpoints:
-            for i, gen in enumerate(self.checkpoints):
-                self.app.ts['what'] = ('removing checkpoint %d/%d' %
-                                        (i+1, len(self.checkpoints)))
+            self.configure_ttystatus_for_checkpoint_removal()
+            for gen in self.checkpoints:
+                self.app.ts['checkpoint'] = gen
                 self.repo.remove_generation(gen)
 
         if self.fs:
@@ -233,15 +384,21 @@ class BackupPlugin(obnamlib.ObnamPlugin):
 
     def make_checkpoint(self):
         logging.info('Making checkpoint')
-        self.app.ts['what'] = 'making chckpoint;'
+        self.app.ts['what'] = 'making checkpoint'
+        self.app.ts.flush()
         if not self.pretend:
             self.checkpoints.append(self.repo.new_generation)
             self.backup_parents('.')
+            self.repo.lock_shared()
+            self.add_chunks_to_shared()
             self.repo.commit_client(checkpoint=True)
+            self.repo.commit_shared()
+            self.last_checkpoint = self.repo.fs.bytes_written
+            self.repo = self.app.open_repository(repofs=self.repo.fs.fs)
             self.repo.lock_client(self.app.settings['client-name'])
             self.repo.start_generation()
-            self.last_checkpoint = self.repo.fs.bytes_written
             self.app.dump_memory_profile('at end of checkpoint')
+        self.app.ts['what'] = self.app.ts['current-file']
 
     def find_files(self, root):
         '''Find all files and directories that need to be backed up.
@@ -252,6 +409,7 @@ class BackupPlugin(obnamlib.ObnamPlugin):
         the directory itself (name, metadata, file list).
         
         '''
+
         for pathname, st in self.fs.scan_tree(root, ok=self.can_be_backed_up):
             tracing.trace('considering %s' % pathname)
             try:
@@ -321,7 +479,8 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                 current.st_nlink != old.st_nlink or
                 current.st_size != old.st_size or
                 current.st_uid != old.st_uid or
-                current.st_gid != old.st_gid)
+                current.st_gid != old.st_gid or
+                current.xattr != old.xattr)
 
     def backup_parents(self, root):
         '''Back up parents of root, non-recursively.'''
@@ -350,14 +509,28 @@ class BackupPlugin(obnamlib.ObnamPlugin):
             tracing.trace('pretending to upload the whole file')
             self.update_progress_with_upload(metadata.st_size)
             return
+
         tracing.trace('setting file chunks to empty')
         if not self.pretend:
             self.repo.set_file_chunks(filename, [])
+
         tracing.trace('opening file for reading')
         f = self.fs.open(filename, 'r')
+
+        summer = self.repo.new_checksummer()
+
+        max_intree = self.app.settings['node-size'] / 4
+        if (metadata.st_size <= max_intree and 
+            self.app.settings['small-files-in-btree']):
+            contents = f.read()
+            assert len(contents) <= max_intree # FIXME: silly error checking
+            f.close()
+            self.repo.set_file_data(filename, contents)
+            summer.update(contents)
+            return summer.digest()
+
         chunk_size = int(self.app.settings['chunk-size'])
         chunkids = []
-        summer = self.repo.new_checksummer()
         while True:
             tracing.trace('reading some data')
             data = f.read(chunk_size)
@@ -374,7 +547,8 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                     self.app.dump_memory_profile('after appending some '
                                                     'chunkids')
                     chunkids = []
-            self.update_progress_with_upload(len(data))
+            else:
+                self.update_progress_with_upload(len(data))
             
             if not self.pretend and self.time_for_checkpoint():
                 logging.debug('making checkpoint in the middle of a file')
@@ -397,13 +571,18 @@ class BackupPlugin(obnamlib.ObnamPlugin):
         '''Back up a chunk of data by putting it into the repository.'''
 
         def find():
-            return self.repo.find_chunks(checksum)
+            return (self.repo.find_chunks(checksum) + 
+                     self.chunkid_pool.get(checksum))
 
         def get(chunkid):
             return self.repo.get_chunk(chunkid)
 
         def put():
-            return self.repo.put_chunk(data, checksum)
+            self.update_progress_with_upload(len(data))
+            return self.repo.put_chunk_only(data)
+            
+        def share(chunkid):
+            self.chunkid_pool.add(chunkid, checksum)
 
         checksum = self.repo.checksum(data)
 
@@ -416,18 +595,24 @@ class BackupPlugin(obnamlib.ObnamPlugin):
                 if data == data2:
                     return chunkid
             else:
-                return put()
+                chunkid = put()
+                share(chunkid)
+                return chunkid
         elif mode == 'fatalist':
             existing = find()
             if existing:
                 return existing[0]
             else:
-                return put()
+                chunkid = put()
+                share(chunkid)
+                return chunkid
         else:
             if not hasattr(self, 'bad_deduplicate_reported'):
                 logging.error('unknown --deduplicate setting value')
                 self.bad_deduplicate_reported = True
-            return put()
+            chunkid = put()
+            share(chunkid)
+            return chunkid
 
     def backup_dir_contents(self, root):
         '''Back up the list of files in a directory.'''
