@@ -23,26 +23,36 @@ import obnamlib
 class EncryptionPlugin(obnamlib.ObnamPlugin):
 
     def enable(self):
+        encryption_group = obnamlib.option_group['encryption'] = 'Encryption'
+
         self.app.settings.string(['encrypt-with'],
                                    'PGP key with which to encrypt data '
-                                        'in the backup repository')
+                                        'in the backup repository',
+                                 group=encryption_group)
         self.app.settings.string(['keyid'],
                                    'PGP key id to add to/remove from '
-                                        'the backup repository')
+                                        'the backup repository',
+                                 group=encryption_group)
         self.app.settings.boolean(['weak-random'],
                                     'use /dev/urandom instead of /dev/random '
-                                        'to generate symmetric keys')
+                                        'to generate symmetric keys',
+                                 group=encryption_group)
         self.app.settings.string(['symmetric-key-bits'],
-                                   'size of symmetric key, in bits')
+                                   'size of symmetric key, in bits',
+                                 group=encryption_group)
         
+        self.tag = "encrypt1"
+
         hooks = [
-            ('repository-toplevel-init', self.toplevel_init),
-            ('repository-read-data', self.toplevel_read_data),
-            ('repository-write-data', self.toplevel_write_data),
-            ('repository-add-client', self.add_client),
+            ('repository-toplevel-init', self.toplevel_init,
+             obnamlib.Hook.DEFAULT_PRIORITY),
+            ('repository-data', self,
+             obnamlib.Hook.LATE_PRIORITY),
+            ('repository-add-client', self.add_client,
+             obnamlib.Hook.DEFAULT_PRIORITY),
         ]
-        for name, callback in hooks:
-            self.app.hooks.add_callback(name, callback)
+        for name, callback, rev in hooks:
+            self.app.hooks.add_callback(name, callback, rev)
             
         self._pubkey = None
         
@@ -82,11 +92,9 @@ class EncryptionPlugin(obnamlib.ObnamPlugin):
 
     def _write_file(self, repo, pathname, contents):
         repo.fs.fs.write_file(pathname, contents)
-        repo.fs.make_readonly(pathname)
 
     def _overwrite_file(self, repo, pathname, contents):
         repo.fs.fs.overwrite_file(pathname, contents)
-        repo.fs.make_readonly(pathname)
 
     def toplevel_init(self, repo, toplevel):
         '''Initialize a new toplevel for encryption.'''
@@ -107,13 +115,13 @@ class EncryptionPlugin(obnamlib.ObnamPlugin):
         encrypted = obnamlib.encrypt_symmetric(encoded, symmetric_key)
         self._write_file(repo, os.path.join(toplevel, 'userkeys'), encrypted)
 
-    def toplevel_read_data(self, encrypted, repo, toplevel):
+    def filter_read(self, encrypted, repo, toplevel):
         if not self.keyid:
             return encrypted
         symmetric_key = self.get_symmetric_key(repo, toplevel)
         return obnamlib.decrypt_symmetric(encrypted, symmetric_key)
 
-    def toplevel_write_data(self, cleartext, repo, toplevel):
+    def filter_write(self, cleartext, repo, toplevel):
         if not self.keyid:
             return cleartext
         symmetric_key = self.get_symmetric_key(repo, toplevel)
@@ -129,12 +137,12 @@ class EncryptionPlugin(obnamlib.ObnamPlugin):
 
     def read_keyring(self, repo, toplevel):
         encrypted = repo.fs.fs.cat(os.path.join(toplevel, 'userkeys'))
-        encoded = self.toplevel_read_data(encrypted, repo, toplevel)
+        encoded = self.filter_read(encrypted, repo, toplevel)
         return obnamlib.Keyring(encoded=encoded)
 
     def write_keyring(self, repo, toplevel, keyring):
         encoded = str(keyring)
-        encrypted = self.toplevel_write_data(encoded, repo, toplevel)
+        encrypted = self.filter_write(encoded, repo, toplevel)
         pathname = os.path.join(toplevel, 'userkeys')
         self._overwrite_file(repo, pathname, encrypted)
 
@@ -179,7 +187,7 @@ class EncryptionPlugin(obnamlib.ObnamPlugin):
         toplevels = repo.fs.listdir('.')
         keys = dict()
         tops = dict()
-        for toplevel in toplevels:
+        for toplevel in [d for d in toplevels if d != 'metadata']:
             userkeys = self.read_keyring(repo, toplevel)
             for keyid in userkeys.keyids():
                 keys[keyid] = keys.get(keyid, []) + [toplevel]
@@ -208,7 +216,7 @@ class EncryptionPlugin(obnamlib.ObnamPlugin):
             for keyid in tops[toplevel]:
                 print '  %s' % keyid
 
-    _shared = ['chunklist', 'chunks', 'chunksums', 'clientlist', 'metadata']
+    _shared = ['chunklist', 'chunks', 'chunksums', 'clientlist']
     
     def _find_clientdirs(self, repo, client_names):
         return [repo.client_dir(repo.clientlist.get_client_id(x))
