@@ -37,10 +37,11 @@ class ForgetPlugin(obnamlib.ObnamPlugin):
         
         self.app.ts['gen'] = None
         self.app.ts['gens'] = []
-        self.app.ts.format('forgetting generation %Index(gen,gens)')
+        self.app.ts.format('forgetting generations: %Index(gen,gens) done')
 
         self.repo = self.app.open_repository()
         self.repo.lock_client(self.app.settings['client-name'])
+        self.repo.lock_shared()
 
         self.app.dump_memory_profile('at beginning')
         if args:
@@ -48,6 +49,7 @@ class ForgetPlugin(obnamlib.ObnamPlugin):
             for genspec in args:
                 self.app.ts['gen'] = genspec
                 genid = self.repo.genspec(genspec)
+                self.app.ts.notify('Forgetting generation %s' % genid)
                 self.remove(genid)
                 self.app.dump_memory_profile('after removing %s' % genid)
         elif self.app.settings['keep']:
@@ -72,11 +74,14 @@ class ForgetPlugin(obnamlib.ObnamPlugin):
                 self.app.dump_memory_profile('after removing %s' % genid)
 
         self.repo.commit_client()
+        self.repo.commit_shared()
         self.app.dump_memory_profile('after committing')
         self.repo.fs.close()
         self.app.ts.finish()
 
     def remove(self, genid):
-        if not self.app.settings['pretend']:
+        if self.app.settings['pretend']:
+            self.app.ts.notify('Pretending to remove generation %s' % genid)
+        else:
             self.repo.remove_generation(genid)
 
