@@ -55,7 +55,7 @@ def ioerror_to_oserror(method):
         try:
             return method(self, filename, *args, **kwargs)
         except IOError, e:
-            raise OSError(e.errno, e.strerror, filename)
+            raise OSError(e.errno, e.strerror or str(e), filename)
     
     return helper
 
@@ -112,26 +112,42 @@ class SftpFS(obnamlib.VirtualFileSystem):
         obnamlib.VirtualFileSystem.__init__(self, baseurl)
         self.sftp = None
         self.settings = settings
+        self._roundtrips = 0
         self.reinit(baseurl, create=create)
 
     def _delay(self):
+        self._roundtrips += 1
         if self.settings:
             ms = self.settings['sftp-delay']
             if ms > 0:
                 time.sleep(ms * 0.001)
+
+    def log_stats(self):
+        obnamlib.VirtualFileSystem.log_stats(self)
+        logging.info('VFS: baseurl=%s roundtrips=%s' %
+                         (self.baseurl, self._roundtrips))
         
     def _to_string(self, str_or_unicode):
         if type(str_or_unicode) is unicode:
             return str_or_unicode.encode('utf-8')
         else:
             return str_or_unicode
+
+    def _create_root_if_missing(self):
+        try:
+            self.mkdir(self.path)
+        except OSError, e:
+            # sftp/paramiko does not give us a useful errno so we hope
+            # for the best
+            pass
+        self.create_path_if_missing = False # only create once
         
     def connect(self):
         try_openssh = not self.settings or not self.settings['pure-paramiko']
         if not try_openssh or not self._connect_openssh():
             self._connect_paramiko()
-        if self.create_path_if_missing and not self.exists(self.path):
-            self.mkdir(self.path)
+        if self.create_path_if_missing:
+            self._create_root_if_missing()
         self.chdir(self.path)
 
     def _connect_openssh(self):
@@ -281,9 +297,8 @@ class SftpFS(obnamlib.VirtualFileSystem):
         self._delay()
         
         if self.sftp:
-            if create and not self.exists(path):
-                self.mkdir(path)
-                self.create_path_if_missing = False # only create first time
+            if create:
+                self._create_root_if_missing()
             self.sftp.chdir(path)
 
     def _get_username(self):
@@ -355,20 +370,16 @@ class SftpFS(obnamlib.VirtualFileSystem):
         fixed = [(name, self._fix_stat(name, st)) for name, st in pairs]
         return fixed
 
-    def lock(self, lockname):
+    def lock(self, lockname, data):
         try:
-            self.write_file(lockname, '')
-        except IOError, e:
-            if e.errno == errno.EEXIST:
-                raise obnamlib.Error('Lock %s already exists' % lockname)
-            else:
-                raise
+            self.write_file(lockname, data)
+        except OSError, e:
+            raise obnamlib.LockFail('Failure get lock %s' % lockname)
 
     def unlock(self, lockname):
         self._remove_if_exists(lockname)
 
     def exists(self, pathname):
-        self._delay()
         try:
             self.lstat(pathname)
         except OSError:
@@ -398,7 +409,6 @@ class SftpFS(obnamlib.VirtualFileSystem):
         
     @ioerror_to_oserror
     def makedirs(self, pathname):
-        self._delay()
         parent = os.path.dirname(pathname)
         if parent and parent != pathname and not self.exists(parent):
             self.makedirs(parent)
@@ -494,85 +504,94 @@ class SftpFS(obnamlib.VirtualFileSystem):
 
     @ioerror_to_oserror
     def write_file(self, pathname, contents):
-        self._delay()
-        self._write_helper(pathname, 'wx', contents)
+        try:
+            f = self.open(pathname, 'wx')
+        except (IOError, OSError), e:
+            if e.errno != errno.ENOENT:
+                raise
+            dirname = os.path.dirname(pathname)
+            self.makedirs(dirname)
+            f = self.open(pathname, 'wx')
+
+        self._write_helper(f, contents)
+        f.close()
 
     def _tempfile(self, dirname):
-        '''Generate a filename that does not exist.
-        
-        This is _not_ as safe as tempfile.mkstemp. Plenty of race
-        conditions. But seems to be as good as SFTP will allow.
-        
-        '''
+        '''Create a new file with a random name, return file handle and name.'''
         
+        if dirname:
+            try:
+                self.makedirs(dirname)
+            except OSError:
+                # We ignore the error, on the assumption that it was due
+                # to the directory already existing. If it didn't exist
+                # and the error was for something else, then we'll catch
+                # that when we open the file for writing.
+                pass
+
         while True:
             i = random.randint(0, 2**64-1)
             basename = 'tmp.%x' % i
             pathname = os.path.join(dirname, basename)
-            if not self.exists(pathname):
-                return pathname
+            try:
+                f = self.open(pathname, 'wx', bufsize=self.chunk_size)
+            except OSError:
+                pass
+            else:
+                return f, pathname
 
     @ioerror_to_oserror
-    def overwrite_file(self, pathname, contents, make_backup=True):
+    def overwrite_file(self, pathname, contents):
         self._delay()
         dirname = os.path.dirname(pathname)
-        tempname = self._tempfile(dirname)
-        self._write_helper(tempname, 'wx', contents)
-
-        # Rename existing to have a .bak suffix. If _that_ file already
-        # exists, remove that.
-        bak = pathname + ".bak"
-        self._remove_if_exists(bak)
-        try:
-            self.rename(pathname, bak)
-        except OSError, e:
-            if e.errno != errno.ENOENT:
-                raise
+        f, tempname = self._tempfile(dirname)
+        self._write_helper(f, contents)
+        f.close()
         self.rename(tempname, pathname)
-        if not make_backup:
-            self._remove_if_exists(bak)
         
-    def _write_helper(self, pathname, mode, contents):
-        self._delay()
-        dirname = os.path.dirname(pathname)
-        if dirname and not self.exists(dirname):
-            self.makedirs(dirname)
-        f = self.open(pathname, mode, bufsize=self.chunk_size)
+    def _write_helper(self, f, contents):
         for pos in range(0, len(contents), self.chunk_size):
             chunk = contents[pos:pos + self.chunk_size]
             f.write(chunk)
             self.bytes_written += len(chunk)
-        f.close()
 
 
 class SftpPlugin(obnamlib.ObnamPlugin):
 
     def enable(self):
+        ssh_group = obnamlib.option_group['ssh'] = 'SSH/SFTP'
+        devel_group = obnamlib.option_group['devel']
+
         self.app.settings.integer(['sftp-delay'],
                                   'add an artificial delay (in milliseconds) '
-                                    'to all SFTP transfers')
+                                    'to all SFTP transfers',
+                                  group=devel_group)
 
         self.app.settings.string(['ssh-key'],
                                  'use FILENAME as the ssh RSA private key for '
                                     'sftp access (default is using keys known '
                                     'to ssh-agent)',
-                                 metavar='FILENAME')
+                                 metavar='FILENAME',
+                                 group=ssh_group)
 
         self.app.settings.boolean(['strict-ssh-host-keys'],
                                   'require that the ssh host key must be '
                                     'known and correct to be accepted; '
-                                    'default is to accept unknown keys')
+                                    'default is to accept unknown keys',
+                                  group=ssh_group)
 
         self.app.settings.string(['ssh-known-hosts'],
                                  'filename of the user\'s known hosts file '
                                     '(default: %default)',
                                  metavar='FILENAME',
                                  default=
-                                    os.path.expanduser('~/.ssh/known_hosts'))
+                                    os.path.expanduser('~/.ssh/known_hosts'),
+                                 group=ssh_group)
 
         self.app.settings.boolean(['pure-paramiko'],
                                  'do not use openssh even if available, '
-                                    'use paramiko only instead')
+                                    'use paramiko only instead',
+                                  group=ssh_group)
 
         self.app.fsf.register('sftp', SftpFS, settings=self.app.settings)
 
