@@ -47,7 +47,6 @@ class HookedFS(object):
         self.repo = repo
         self.fs = fs
         self.hooks = hooks
-        self.do_make_readonly = True
         
     def __getattr__(self, name):
         return getattr(self.fs, name)
@@ -59,38 +58,32 @@ class HookedFS(object):
         else: # pragma: no cover
             raise obnamlib.Error('File at repository root: %s' % filename)
         
-    def cat(self, filename):
+    def cat(self, filename, runfilters=True):
         data = self.fs.cat(filename)
         toplevel = self._get_toplevel(filename)
-        return self.hooks.call('repository-read-data', data,
-                                repo=self.repo, toplevel=toplevel)
-        
-    def write_file(self, filename, data):
+        if not runfilters:
+            return data
+        return self.hooks.filter_read('repository-data', data,
+                                      repo=self.repo, toplevel=toplevel)
+
+    def lock(self, filename, data):
+        self.fs.lock(filename, data)
+
+    def write_file(self, filename, data, runfilters=True):
         tracing.trace('writing hooked %s' % filename)
         toplevel = self._get_toplevel(filename)
-        data = self.hooks.call('repository-write-data', data,
-                                repo=self.repo, toplevel=toplevel)
+        if runfilters:
+            data = self.hooks.filter_write('repository-data', data,
+                                           repo=self.repo, toplevel=toplevel)
         self.fs.write_file(filename, data)
-        self.make_readonly(filename)
         
-    def overwrite_file(self, filename, data):
+    def overwrite_file(self, filename, data, runfilters=True):
         tracing.trace('overwriting hooked %s' % filename)
         toplevel = self._get_toplevel(filename)
-        data = self.hooks.call('repository-write-data', data,
-                                repo=self.repo, toplevel=toplevel)
+        if runfilters:
+            data = self.hooks.filter_write('repository-data', data,
+                                           repo=self.repo, toplevel=toplevel)
         self.fs.overwrite_file(filename, data)
-        self.make_readonly(filename)
-        
-    def make_readonly(self, filename):
-        if self.do_make_readonly:
-            try:
-                self.fs.chmod(filename, stat.S_IRUSR)
-            except OSError, e: # pragma: no cover
-                if e.errno != errno.EPERM:
-                    raise
-                else:
-                    logging.debug('Ignoring chmod error in repository')
-                    self.do_make_readonly = False
         
 
 class Repository(object):
@@ -126,10 +119,11 @@ class Repository(object):
 
     '''
     
-    format_version = 5
+    format_version = 6
 
     def __init__(self, fs, node_size, upload_queue_size, lru_size, hooks,
-                 idpath_depth, idpath_bits, idpath_skip, current_time):
+                 idpath_depth, idpath_bits, idpath_skip, current_time,
+                 lock_timeout, client_name):
 
         self.current_time = current_time
         self.setup_hooks(hooks or obnamlib.HookManager())
@@ -137,12 +131,17 @@ class Repository(object):
         self.node_size = node_size
         self.upload_queue_size = upload_queue_size
         self.lru_size = lru_size
+        
+        hider = hashlib.md5()
+        hider.update(client_name)
+
+        self.lockmgr = obnamlib.LockManager(self.fs, lock_timeout, 
+                                            hider.hexdigest())
+
         self.got_root_lock = False
-        self.clientlist = obnamlib.ClientList(self.fs, node_size, 
-                                              upload_queue_size, 
-                                              lru_size, self)
+        self._open_client_list()
+        self.got_shared_lock = False
         self.got_client_lock = False
-        self.client_lockfile = None
         self.current_client = None
         self.current_client_id = None
         self.new_generation = None
@@ -150,23 +149,32 @@ class Repository(object):
         self.removed_clients = []
         self.removed_generations = []
         self.client = None
-        self.chunklist = obnamlib.ChunkList(self.fs, node_size, 
-                                            upload_queue_size, 
-                                            lru_size, self)
-        self.chunksums = obnamlib.ChecksumTree(self.fs, 'chunksums', 
-                                               len(self.checksum('')),
-                                               node_size, upload_queue_size, 
-                                               lru_size, self)
+        self._open_shared()
         self.prev_chunkid = None
         self.chunk_idpath = larch.IdPath('chunks', idpath_depth, 
                                          idpath_bits, idpath_skip)
+        self._chunks_exists = False
+
+    def _open_client_list(self):
+        self.clientlist = obnamlib.ClientList(self.fs, self.node_size, 
+                                              self.upload_queue_size, 
+                                              self.lru_size, self)
+
+    def _open_shared(self):
+        self.chunklist = obnamlib.ChunkList(self.fs, self.node_size, 
+                                            self.upload_queue_size, 
+                                            self.lru_size, self)
+        self.chunksums = obnamlib.ChecksumTree(self.fs, 'chunksums', 
+                                               len(self.checksum('')),
+                                               self.node_size, 
+                                               self.upload_queue_size, 
+                                               self.lru_size, self)
 
     def setup_hooks(self, hooks):
         self.hooks = hooks
         
         self.hooks.new('repository-toplevel-init')
-        self.hooks.new_filter('repository-read-data')
-        self.hooks.new_filter('repository-write-data')
+        self.hooks.new_filter('repository-data')
         self.hooks.new('repository-add-client')
         
     def checksum(self, data):
@@ -174,10 +182,6 @@ class Repository(object):
         
         The checksum is (currently) MD5.
         
-        Return a non-binary string (hexdigest) form of the checksum
-        so that it can easily be used for filenames, or printed to
-        log files, or whatever.
-        
         '''
 
         checksummer = self.new_checksummer()
@@ -211,10 +215,15 @@ class Repository(object):
         if not self.got_root_lock:
             raise LockFail('have not got lock on root node')
 
+    def require_shared_lock(self):
+        '''Ensure we have the lock on the shared B-trees except clientlist.'''
+        if not self.got_shared_lock:
+            raise LockFail('have not got lock on shared B-trees')
+
     def require_client_lock(self):
         '''Ensure we have the lock on the currently open client.'''
         if not self.got_client_lock:
-                raise LockFail('have not got lock on client')
+            raise LockFail('have not got lock on client')
 
     def require_open_client(self):
         '''Ensure we have opened the client (r/w or r/o).'''
@@ -226,6 +235,21 @@ class Repository(object):
         if self.new_generation is None:
             raise obnamlib.Error('new generation has not started')
 
+    def require_no_root_lock(self):
+        '''Ensure we haven't locked root yet.'''
+        if self.got_root_lock:
+            raise obnamlib.Error('We have already locked root, oops')
+
+    def require_no_shared_lock(self):
+        '''Ensure we haven't locked shared B-trees yet.'''
+        if self.got_shared_lock:
+            raise obnamlib.Error('We have already locked shared B-trees, oops')
+
+    def require_no_client_lock(self):
+        '''Ensure we haven't locked the per-client B-tree yet.'''
+        if self.got_client_lock:
+            raise obnamlib.Error('We have already locked the client, oops')
+
     def lock_root(self):
         '''Lock root node.
         
@@ -234,17 +258,18 @@ class Repository(object):
         
         '''
 
-        tracing.trace('locking root')        
+        tracing.trace('locking root')
+        self.require_no_root_lock()
+        self.require_no_client_lock()
+        self.require_no_shared_lock()
+
+        self.lockmgr.lock(['.'])
         self.check_format_version()
-        try:
-            self.fs.fs.write_file('root.lock', '')
-        except OSError, e:
-            if e.errno == errno.EEXIST:
-                raise LockFail('Lock file root.lock already exists')
         self.got_root_lock = True
         self.added_clients = []
         self.removed_clients = []
         self._write_format_version(self.format_version)
+        self.clientlist.start_changes()
 
     def unlock_root(self):
         '''Unlock root node without committing changes made.'''
@@ -252,8 +277,9 @@ class Repository(object):
         self.require_root_lock()
         self.added_clients = []
         self.removed_clients = []
-        self.fs.remove('root.lock')
+        self.lockmgr.unlock(['.'])
         self.got_root_lock = False
+        self._open_client_list()
         
     def commit_root(self):
         '''Commit changes to root node, and unlock it.'''
@@ -281,7 +307,7 @@ class Repository(object):
         '''
         
         if self.fs.exists('metadata/format'):
-            data = self.fs.cat('metadata/format')
+            data = self.fs.cat('metadata/format', runfilters=False)
             lines = data.splitlines()
             line = lines[0]
             try:
@@ -300,8 +326,8 @@ class Repository(object):
         tracing.trace('write format version')
         if not self.fs.exists('metadata'):
             self.fs.mkdir('metadata')
-            self.hooks.call('repository-toplevel-init', self, 'metadata')
-        self.fs.overwrite_file('metadata/format', '%s\n' % version)
+        self.fs.overwrite_file('metadata/format', '%s\n' % version,
+                               runfilters=False)
 
     def check_format_version(self):
         '''Verify that on-disk format version is compatbile.
@@ -312,8 +338,9 @@ class Repository(object):
         
         on_disk = self.get_format_version()
         if on_disk is not None and not self.acceptable_version(on_disk):
-            raise BadFormat('On-disk format %s is incompatible '
-                            'with program format %s' %
+            raise BadFormat('On-disk repository format %s is incompatible '
+                            'with program format %s; you need to use a '
+                            'different version of Obnam' %
                                 (on_disk, self.format_version))
         
     def add_client(self, client_name):
@@ -338,6 +365,53 @@ class Repository(object):
         if client_name not in self.list_clients():
             raise obnamlib.Error('client %s does not exist' % client_name)
         self.removed_clients.append(client_name)
+
+    @property
+    def shared_dirs(self):
+        return [self.chunklist.dirname, self.chunksums.dirname,
+                self.chunk_idpath.dirname]
+        
+    def lock_shared(self):
+        '''Lock a client for exclusive write access.
+        
+        Raise obnamlib.LockFail if locking fails. Lock will be released
+        by commit_client() or unlock_client().
+
+        '''
+
+        tracing.trace('locking shared')
+        self.require_no_shared_lock()
+        self.check_format_version()
+        self.lockmgr.lock(self.shared_dirs)
+        self.got_shared_lock = True
+        tracing.trace('starting changes in chunksums and chunklist')
+        self.chunksums.start_changes()
+        self.chunklist.start_changes()
+        
+        # Initialize the chunks directory for encryption, etc, if it just
+        # got created.
+        dirname = self.chunk_idpath.dirname
+        filenames = self.fs.listdir(dirname)
+        if filenames == [] or filenames == ['lock']:
+            self.hooks.call('repository-toplevel-init', self, dirname)
+
+
+    def commit_shared(self):
+        '''Commit changes to shared B-trees.'''
+        
+        tracing.trace('committing shared')
+        self.require_shared_lock()
+        self.chunklist.commit()
+        self.chunksums.commit()
+        self.unlock_shared()
+
+    def unlock_shared(self):
+        '''Unlock currently locked shared B-trees.'''
+        tracing.trace('unlocking shared')
+        self.require_shared_lock()
+        self.lockmgr.unlock(self.shared_dirs)
+        self.got_shared_lock = False
+        self._open_shared()
         
     def lock_client(self, client_name):
         '''Lock a client for exclusive write access.
@@ -348,6 +422,9 @@ class Repository(object):
         '''
 
         tracing.trace('client_name=%s', client_name)
+        self.require_no_client_lock()
+        self.require_no_shared_lock()
+        
         self.check_format_version()
         client_id = self.clientlist.get_client_id(client_name)
         if client_id is None:
@@ -357,14 +434,9 @@ class Repository(object):
         if not self.fs.exists(client_dir):
             self.fs.mkdir(client_dir)
             self.hooks.call('repository-toplevel-init', self, client_dir)
-        lockname = os.path.join(client_dir, 'lock')
-        try:
-            self.fs.write_file(lockname, '')
-        except OSError, e:
-            if e.errno == errno.EEXIST:
-                raise LockFail('client %s is already locked' % client_name)
+
+        self.lockmgr.lock([client_dir])
         self.got_client_lock = True
-        self.client_lockfile = lockname
         self.current_client = client_name
         self.current_client_id = client_id
         self.added_generations = []
@@ -380,13 +452,11 @@ class Repository(object):
         tracing.trace('unlocking client')
         self.require_client_lock()
         self.new_generation = None
-        for genid in self.added_generations:
-            self._really_remove_generation(genid)
+        self._really_remove_generations(self.added_generations)
+        self.lockmgr.unlock([self.client.dirname])
         self.client = None # FIXME: This should remove uncommitted data.
         self.added_generations = []
         self.removed_generations = []
-        self.fs.remove(self.client_lockfile)
-        self.client_lockfile = None
         self.got_client_lock = False
         self.current_client = None
         self.current_client_id = None
@@ -395,14 +465,14 @@ class Repository(object):
         '''Commit changes to and unlock currently locked client.'''
         tracing.trace('committing client (checkpoint=%s)', checkpoint)
         self.require_client_lock()
+        self.require_shared_lock()
+        commit_client = self.new_generation or self.removed_generations
         if self.new_generation:
             self.client.set_current_generation_is_checkpoint(checkpoint)
         self.added_generations = []
-        for genid in self.removed_generations:
-            self._really_remove_generation(genid)
-        self.client.commit()
-        self.chunklist.commit()
-        self.chunksums.commit()
+        self._really_remove_generations(self.removed_generations)
+        if commit_client:
+            self.client.commit()
         self.unlock_client()
         
     def open_client(self, client_name):
@@ -448,8 +518,8 @@ class Repository(object):
         self.added_generations.append(self.new_generation)
         return self.new_generation
 
-    def _really_remove_generation(self, gen_id):
-        '''Really remove a committed generation.
+    def _really_remove_generations(self, remove_genids):
+        '''Really remove a list of generations.
         
         This is not part of the public API.
         
@@ -457,31 +527,50 @@ class Repository(object):
         
         '''
 
+        def find_chunkids_in_gens(genids):
+            chunkids = set()
+            for genid in genids:
+                x = self.client.list_chunks_in_generation(genid)
+                chunkids = chunkids.union(set(x))
+            return chunkids
 
-        def filter_away_chunks_used_by_other_gens(chunk_ids, gen_id):
-            for other_id in self.list_generations():
-                if other_id != gen_id:
-                    other_chunks = self.client.list_chunks_in_generation(
-                                        other_id)
-                    other_chunks = set(other_chunks)
-                    chunk_ids = chunk_ids.difference(other_chunks)
+        def find_gens_to_keep():
+            return [genid
+                    for genid in self.list_generations()
+                    if genid not in remove_genids]
 
-            return chunk_ids
-
-        def remove_unused_chunks(chunk_ids):
+        def remove_chunks(chunk_ids):
             for chunk_id in chunk_ids:
-                checksum = self.chunklist.get_checksum(chunk_id)
-                self.chunksums.remove(checksum, chunk_id, 
-                                      self.current_client_id)
-                if not self.chunksums.chunk_is_used(checksum, chunk_id):
+                try:
+                    checksum = self.chunklist.get_checksum(chunk_id)
+                except KeyError:
+                    # No checksum, therefore it can't be shared, therefore
+                    # we can remove it.
                     self.remove_chunk(chunk_id)
+                else:
+                    self.chunksums.remove(checksum, chunk_id, 
+                                          self.current_client_id)
+                    if not self.chunksums.chunk_is_used(checksum, chunk_id):
+                        self.remove_chunk(chunk_id)
+
+        def remove_gens(genids):
+            if self.new_generation is None:
+                self.client.start_changes(create_tree=False)
+            for genid in genids:
+                self.client.remove_generation(genid)
+
+        if not remove_genids:
+            return
 
         self.require_client_lock()
-        logging.debug('_really_remove_generation: %d' % gen_id)
-        chunk_ids = set(self.client.list_chunks_in_generation(gen_id))
-        chunk_ids = filter_away_chunks_used_by_other_gens(chunk_ids, gen_id)
-        remove_unused_chunks(chunk_ids)
-        self.client.remove_generation(gen_id)
+        self.require_shared_lock()
+
+        maybe_remove = find_chunkids_in_gens(remove_genids)
+        keep_genids = find_gens_to_keep()
+        keep = find_chunkids_in_gens(keep_genids)
+        remove = maybe_remove.difference(keep)
+        remove_chunks(remove)
+        remove_gens(remove_genids)
 
     def remove_generation(self, gen):
         '''Remove a committed generation.'''
@@ -529,16 +618,9 @@ class Repository(object):
     def _chunk_filename(self, chunkid):
         return self.chunk_idpath.convert(chunkid)
 
-    def put_chunk(self, data, checksum):
+    def put_chunk_only(self, data):
         '''Put chunk of data into repository.
         
-        checksum is the checksum of the data, and must be the same
-        value as returned by self.checksum(data). However, since all
-        known use cases require the caller to know the checksum before
-        calling this method, and since computing checksums is
-        expensive, we micro-optimize a little bit by passing it as
-        an argument.
-        
         If the same data is already in the repository, it will be put there
         a second time. It is the caller's responsibility to check
         that the data is not already in the repository.
@@ -550,29 +632,45 @@ class Repository(object):
         def random_chunkid():
             return random.randint(0, obnamlib.MAX_ID)
         
-        tracing.trace('putting chunk (checksum=%s)', repr(checksum))
         self.require_started_generation()
+
         if self.prev_chunkid is None:
             self.prev_chunkid = random_chunkid()
+
         while True:
             chunkid = (self.prev_chunkid + 1) % obnamlib.MAX_ID
             filename = self._chunk_filename(chunkid)
-            if not self.fs.exists(filename):
+            try:
+                self.fs.write_file(filename, data)
+            except OSError, e: # pragma: no cover
+                if e.errno == errno.EEXIST:
+                    self.prev_chunkid = random_chunkid()
+                    continue
+                raise
+            else:
+                tracing.trace('chunkid=%s', chunkid)
                 break
-            self.prev_chunkid = random_chunkid() # pragma: no cover
-        tracing.trace('chunkid=%s', chunkid)
+
         self.prev_chunkid = chunkid
-        if not self.fs.exists('chunks'):
-            self.fs.mkdir('chunks')
-            self.hooks.call('repository-toplevel-init', self, 'chunks')
-        dirname = os.path.dirname(filename)
-        if not self.fs.exists(dirname):
-            self.fs.makedirs(dirname)
-        self.fs.write_file(filename, data)
-        checksum = self.checksum(data)
+        return chunkid
+
+    def put_chunk_in_shared_trees(self, chunkid, checksum):
+        '''Put the chunk into the shared trees.
+        
+        The chunk is assumed to already exist in the repository, so we
+        just need to add it to the shared trees that map chunkids to
+        checksums and checksums to chunkids.
+        
+        '''
+
+        tracing.trace('chunkid=%s', chunkid)
+        tracing.trace('checksum=%s', repr(checksum))
+
+        self.require_started_generation()
+        self.require_shared_lock()
+
         self.chunklist.add(chunkid, checksum)
         self.chunksums.add(checksum, chunkid, self.current_client_id)
-        return chunkid
         
     def get_chunk(self, chunkid):
         '''Return data of chunk with given id.'''
@@ -618,6 +716,7 @@ class Repository(object):
 
         tracing.trace('chunk_id=%s', chunk_id)
         self.require_open_client()
+        self.require_shared_lock()
         self.chunklist.remove(chunk_id)
         filename = self._chunk_filename(chunk_id)
         try:
@@ -649,6 +748,16 @@ class Repository(object):
         
         self.require_started_generation()
         self.client.append_file_chunks(filename, chunkids)
+
+    def set_file_data(self, filename, contents): # pragma: no cover
+        '''Store contents of file in B-tree instead of chunks dir.'''
+        self.require_started_generation()
+        self.client.set_file_data(filename, contents)
+
+    def get_file_data(self, gen, filename): # pragma: no cover
+        '''Returned contents of file stored in B-tree instead of chunks dir.'''
+        self.require_open_client()
+        return self.client.get_file_data(gen, filename)
         
     def genspec(self, spec):
         '''Interpret a generation specification.'''
