@@ -89,7 +89,7 @@ class VirtualFileSystem(object):
     def listdir2(self, pathname):
         '''Return list of basenames and stats of entities at pathname.'''
 
-    def lock(self, lockname):
+    def lock(self, lockname, data):
         '''Create a lock file with the given name.'''
 
     def unlock(self, lockname):
@@ -185,24 +185,20 @@ class VirtualFileSystem(object):
     def write_file(self, pathname, contents):
         '''Write a new file.
 
-        The file must not yet exist. The file is written atomically,
-        so that the given name will only exist when the file is
-        completely written.
+        The file must not yet exist. The file is not necessarily written 
+        atomically, meaning that if the writing fails (connection to
+        server drops, for example), the file might exist in a partial
+        form. The callers need to deal with this.
         
         Any directories in pathname will be created if necessary.
 
         '''
 
-    def overwrite_file(self, pathname, contents, make_backup=True):
-        '''Like write_file, but overwrites existing file.
+    def overwrite_file(self, pathname, contents):
+        '''Like write_file, but overwrites existing file.'''
 
-        The old file isn't immediately lost, it gets renamed with
-        a backup suffix. The backup file is removed if make_backup is
-        set to False (default is True).
-
-        '''
-
-    def scan_tree(self, dirname, ok=None, dirst=None, log=logging.error):
+    def scan_tree(self, dirname, ok=None, dirst=None, log=logging.error,
+                  error_handler=None):
         '''Scan a tree for files.
         
         Return a generator that returns ``(pathname, stat_result)``
@@ -222,20 +218,27 @@ class VirtualFileSystem(object):
         
         Errors from calling ``listdir`` or ``lstat`` are logged,
         but do not stop the scanning. Such files or directories are
-        not returned, however.
+        not returned, however. If `error_handler` is defined, it is
+        called once for every problem, giving the name and exception
+        as arguments.
         
         '''
+        
+        error_handler = error_handler or (lambda name, e: None)
 
         try:
             pairs = self.listdir2(dirname)
         except OSError, e:
             log('listdir failed: %s: %s' % (e.filename, e.strerror))
+            error_handler(dirname, e)
             pairs = []
             
         queue = []
         for name, st in pairs:
             pathname = os.path.join(dirname, name)
-            if ok is None or ok(pathname, st):
+            if isinstance(st, BaseException):
+                error_handler(pathname, st)
+            elif ok is None or ok(pathname, st):
                 if stat.S_ISDIR(st.st_mode):
                     for t in self.scan_tree(pathname, ok=ok, dirst=st):
                         yield t
@@ -367,23 +370,17 @@ class VfsTests(object): # pragma: no cover
         self.assertEqual(self.fs.getcwd(), self.basepath)
 
     def test_creates_lock_file(self):
-        self.fs.lock('lock')
-        self.assert_(self.fs.exists('lock'))
+        self.fs.lock('lock', 'lock data')
+        self.assertTrue(self.fs.exists('lock'))
+        self.assertEqual(self.fs.cat('lock'), 'lock data')
 
     def test_second_lock_fails(self):
-        self.fs.lock('lock')
-        self.assertRaises(Exception, self.fs.lock, 'lock')
-
-    def test_lock_raises_oserror_without_eexist(self):
-        def raise_it(relative_path, contents):
-            e = OSError()
-            e.errno = errno.EAGAIN
-            raise e
-        self.fs.write_file = raise_it
-        self.assertRaises(OSError, self.fs.lock, 'foo')
+        self.fs.lock('lock', 'lock data')
+        self.assertRaises(Exception, self.fs.lock, 'lock', 'second lock')
+        self.assertEqual(self.fs.cat('lock'), 'lock data')
 
     def test_unlock_removes_lock(self):
-        self.fs.lock('lock')
+        self.fs.lock('lock', 'lock data')
         self.fs.unlock('lock')
         self.assertFalse(self.fs.exists('lock'))
 
@@ -610,26 +607,6 @@ class VfsTests(object): # pragma: no cover
         self.fs.overwrite_file('foo', 'bar')
         self.assertEqual(self.fs.cat('foo'), 'bar')
 
-    def test_overwrite_renames_existing_file(self):
-        self.fs.write_file('foo', 'bar')
-        self.fs.overwrite_file('foo', 'foobar')
-        self.assert_(self.fs.exists('foo.bak'))
-
-    def test_overwrite_removes_existing_bak_file(self):
-        self.fs.write_file('foo', 'bar')
-        self.fs.write_file('foo.bak', 'baz')
-        self.fs.overwrite_file('foo', 'foobar')
-        self.assertEqual(self.fs.cat('foo.bak'), 'bar')
-
-    def test_overwrite_removes_bak_file(self):
-        self.fs.write_file('foo', 'bar')
-        self.fs.overwrite_file('foo', 'foobar', make_backup=False)
-        self.assertFalse(self.fs.exists('foo.bak'))
-
-    def test_overwrite_is_ok_without_bak(self):
-        self.fs.overwrite_file('foo', 'foobar', make_backup=False)
-        self.assertFalse(self.fs.exists('foo.bak'))
-
     def test_overwrite_replaces_existing_file(self):
         self.fs.write_file('foo', 'bar')
         self.fs.overwrite_file('foo', 'foobar')
